@@ -42,8 +42,13 @@ typedef unsigned long long u64;
 #include "nbd.h"
 #endif
 #ifdef NBD_H_LINUX
-#include <linux/types.h>
-#include <linux/nbd.h>
+# ifdef HAVE_LINUX_TYPES_H
+#  include <linux/types.h>
+# else
+#  define __be32 u32
+#  define __be64 u64
+# endif
+# include <linux/nbd.h>
 #endif
 
 #if NBD_LFS==1
@@ -74,6 +79,16 @@ void setmysockopt(int sock) {
 #endif
 }
 
+#ifndef G_GNUC_NORETURN
+#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4)
+#define G_GNUC_NORETURN __attribute__((__noreturn__))
+#else
+#define G_GNUC_NORETURN
+#endif
+#endif
+
+void err(const char *s) G_GNUC_NORETURN;
+
 void err(const char *s) {
 	const int maxlen = 150;
 	char s1[maxlen], *s2;
@@ -96,9 +111,8 @@ void err(const char *s) {
 	s1[maxlen-1] = '\0';
 #ifdef ISSERVER
 	syslog(LOG_ERR, "%s", s1);
-#else
-	fprintf(stderr, "Error: %s\n", s1);
 #endif
+	fprintf(stderr, "Error: %s\n", s1);
 	exit(1);
 }
 
@@ -124,3 +138,7 @@ u64 ntohll(u64 a) {
 }
 #endif
 #define htonll ntohll
+
+/* Flags used between the client and server */
+#define NBD_FLAG_HAS_FLAGS	(1 << 0)	/* Flags are there */
+#define NBD_FLAG_READ_ONLY	(1 << 1)	/* Device is read-only */
