@@ -28,6 +28,7 @@
 #include <syslog.h>
 #include <stdlib.h>
 #include <sys/mount.h>
+#include <sys/mman.h>
 #include <errno.h>
 
 #ifndef __GNUC__
@@ -40,11 +41,16 @@
 
 int check_conn(char* devname, int do_print) {
 	char buf[256];
+	char* p;
 	int fd;
 	int len;
 	if(!strncmp(devname, "/dev/", 5)) {
 		devname+=5;
 	}
+	if((p=strchr(devname, 'p'))) {
+		/* We can't do checks on partitions. */
+		*p='\0';
+	}
 	snprintf(buf, 256, "/sys/block/%s/pid", devname);
 	if((fd=open(buf, O_RDONLY))<0) {
 		if(errno==ENOENT) {
@@ -172,27 +178,23 @@ void setsizes(int nbd, u64 size64, int blocksize, u32 flags) {
 }
 
 void set_timeout(int nbd, int timeout) {
-#ifdef NBD_SET_TIMEOUT
 	if (timeout) {
+#ifdef NBD_SET_TIMEOUT
 		if (ioctl(nbd, NBD_SET_TIMEOUT, (unsigned long)timeout) < 0)
 			err("Ioctl NBD_SET_TIMEOUT failed: %m\n");
 		fprintf(stderr, "timeout=%d\n", timeout);
-	}
+#else
+		err("Ioctl NBD_SET_TIMEOUT cannot be called when compiled on a system that does not support it\n");
 #endif
+	}
 }
 
 void finish_sock(int sock, int nbd, int swap) {
 	if (ioctl(nbd, NBD_SET_SOCK, sock) < 0)
 		err("Ioctl NBD_SET_SOCK failed: %m\n");
 
-#ifndef SO_SWAPPING
 	if (swap)
-		err("You have to compile me on machine with swapping patch enabled in order to use it later.");
-#else
-	if (swap)
-		if (setsockopt(sock, SOL_SOCKET, SO_SWAPPING, &one, sizeof(int)) < 0)
-			err("Could not enable swapping: %m");
-#endif
+		mlockall(MCL_CURRENT | MCL_FUTURE);
 }
 
 int main(int argc, char *argv[]) {
@@ -203,6 +205,7 @@ int main(int argc, char *argv[]) {
 	int cont=0;
 	int timeout=0;
 	int sdp=0;
+	int nofork=0;
 	u64 size64;
 	u32 flags;
 
@@ -211,7 +214,7 @@ int main(int argc, char *argv[]) {
 	if (argc < 3) {
 	errmsg:
 		fprintf(stderr, "nbd-client version %s\n", PACKAGE_VERSION);
-		fprintf(stderr, "Usage: nbd-client [bs=blocksize] [timeout=sec] host port nbd_device [-swap] [-persist]\n");
+		fprintf(stderr, "Usage: nbd-client [bs=blocksize] [timeout=sec] host port nbd_device [-swap] [-persist] [-nofork]\n");
 		fprintf(stderr, "Or   : nbd-client -d nbd_device\n");
 		fprintf(stderr, "Or   : nbd-client -c nbd_device\n");
 		fprintf(stderr, "Default value for blocksize is 1024 (recommended for ethernet)\n");
@@ -222,11 +225,11 @@ int main(int argc, char *argv[]) {
 	}
 
 	++argv; --argc; /* skip programname */
-	
+
 	if (strcmp(argv[0], "-d")==0) {
 		nbd = open(argv[1], O_RDWR);
 		if (nbd < 0)
-			err("Can not open NBD: %m");
+			err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
 		printf("Disconnecting: que, ");
 		if (ioctl(nbd, NBD_CLEAR_QUE)< 0)
 			err("Ioctl failed: %m\n");
@@ -270,7 +273,7 @@ int main(int argc, char *argv[]) {
 	nbddev = argv[0];
 	nbd = open(nbddev, O_RDWR);
 	if (nbd < 0)
-	  err("Can not open NBD: %m");
+	  err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
 	++argv; --argc; /* skip device */
 
 	if (argc>3) goto errmsg;
@@ -292,6 +295,12 @@ int main(int argc, char *argv[]) {
 			++argv;--argc;
 		}
 	}
+	if (argc) {
+		if(strncmp(argv[0], "-nofork", 7)==0) {
+			nofork=1;
+			++argv;--argc;
+		}
+	}
 	if(argc) goto errmsg;
 	sock = opennet(hostname, port, sdp);
 	argv=NULL; argc=0; /* don't use it later suddenly */
@@ -303,10 +312,21 @@ int main(int argc, char *argv[]) {
 
 	/* Go daemon */
 	
-	chdir("/");
-	do {
 #ifndef NOFORK
+	if(!nofork) daemon(0,0);
+	do {
 		if (fork()) {
+			/* Due to a race, the kernel NBD driver cannot
+			 * call for a reread of the partition table
+			 * in the handling of the NBD_DO_IT ioctl().
+			 * Therefore, this is done in the first open()
+			 * of the device. We therefore make sure that
+			 * the device is opened at least once after the
+			 * connection was made. This has to be done in a
+			 * separate process, since the NBD_DO_IT ioctl()
+			 * does not return until the NBD device has
+			 * disconnected.
+			 */
 			while(check_conn(nbddev, 0)) {
 				sleep(1);
 			}
